from subprocess import PIPE, STDOUT, CalledProcessError, Popen
from threading import Thread


class ProcessesContextManager:
    """A context manager that kills any processes given to it on exit from its context."""

    def __init__(self, processes):
        self.processes = processes

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        for process in self.processes:
            process.kill()


def run_logged_subprocess(command, logger, log_level="info", *args, **kwargs):
    """Run a subprocess, sending its stdout and stderr output to the given logger. Extra `args` and `kwargs` are
    provided to the `subprocess.Popen` instance used.

    :param iter(str) command: command to run
    :param logging.Logger logger: logger to use to log stdout and stderr
    :param str log_level: level to log output at
    :raise CalledProcessError: if the subprocess fails (i.e. if it doesn't exit with a 0 return code)
    :return subprocess.CompletedProcess:
    """

    def _log_lines_from_stream(stream, logger):
        """Log lines from the given stream.

        :param io.BufferedReader stream:
        :param logging.Logger logger:
        :return None:
        """
        with stream:
            for line in iter(stream.readline, b""):
                getattr(logger, log_level.lower())(line.decode().strip())

    process = Popen(command, stdout=PIPE, stderr=STDOUT, *args, **kwargs)
    Thread(target=_log_lines_from_stream, args=[process.stdout, logger]).start()
    process.wait()

    if process.returncode != 0:
        raise CalledProcessError(returncode=process.returncode, cmd=" ".join(command))

    return process



import logging
import os

from cleaner import clean, read_csv_files, read_dat_file

from octue.resources import Datafile, Dataset
from octue.utils.files import RegisteredTemporaryDirectory
from tests import TEST_BUCKET_NAME


logger = logging.getLogger(__name__)


def run(analysis):
    """Read a time series of files from a dataset, clean them, and write a new, cleaned dataset.

    See the "fractal" template for an introduction to the analysis object and the purpose of this 'run' function.

    Here, let's create an example application designed to clean up CSV data files produced by an instrument, in this
    case a meteorological mast.

    The aim of this example is to teach you how to use input and output file manifests in an app - so what we'll do is:
    - Use the input manifest to read a sequence of files
    - Perform a simple transformation on some of the data (as if we were doing a data cleaning process)
    - Create new files containing the cleaned data
    - Add them to the output manifest

    :param octue.resources.Analysis analysis:
    :return None:
    """
    # You can use a logger to record debug statements, general information, warnings or errors.
    logger.info("Starting clean up of files in %s", analysis.input_manifest)

    # Get the configuration value for our time averaging window (or if not present, use the default specified in
    # the twine).
    time_window = analysis.configuration_values.get("time_window", 600)
    logger.info("Averaging window set to %ss", time_window)

    # Get the input dataset which will be read in.
    input_dataset = analysis.input_manifest.get_dataset("raw_met_mast_data")

    # There are two types of files in the dataset. Metadata file(s), saved daily, and measurement files (saved hourly).
    # Because a manifest has been created, we're able to get this data out easily with the dataset filtering
    # capabilities. Let's get the metadata and the timeseries files, whilst showing off a couple of the filters.
    # See the Dataset class help for more.
    metadata_file = input_dataset.get_file_by_label("meta")

    timeseries_files = input_dataset.files.filter(labels__contains="timeseries").order_by(
        "tags__sequence",
        check_start_value=0,
        check_constant_increment=1,
    )

    # We used these because they're special helpers - in this case ensuring that there's only one metadata file and
    # ensuring that the timeseries files come in a strictly ordered sequence.
    #
    # We could also have picked up one or more files using general filters, like so:
    #
    #    metadata_files = input_dataset.files.filter(name__icontains="meta")
    #
    # There's generally a few ways to do it. Choose one which is likely to be most consistent - for example if your
    # filenames might be subject to change, but you have better control over the labels, rely on those.

    # At this point it's over to you, to do whatever you want with the contents of these files.
    # For this example app, we will:

    # Use a custom function to read in the strange metadata file that came with the dataset.
    metadata = read_dat_file(metadata_file)

    # Read the sequence of CSV files and concatenate into a pandas dataframe (like a table).
    data = read_csv_files(timeseries_files)

    # Clean the timeseries data up.
    data = clean(data, metadata["date"])

    # Create a temporary directory for the output dataset. This avoids any race conditions arising (if other instances
    # of this application are running at the same time) and avoids any data loss due to overwriting. The temporary
    # directory is deleted once the "with" block is exited.
    temporary_directory = RegisteredTemporaryDirectory().name

    timeseries_datafile = Datafile(
        path=os.path.join(temporary_directory, "cleaned.csv"),
        labels=["timeseries"],
    )

    # Write the file (now we know where to write it)
    with timeseries_datafile.open("w") as fp:
        data.to_csv(path_or_buf=fp)

    # You can replace empty output datasets with datasets instantiated from a local or cloud directory.
    analysis.output_manifest.datasets["cleaned_met_mast_data"] = Dataset(
        path=temporary_directory,
        name="cleaned_met_mast_data",
    )

    # We'll add some labels, which will help to improve searchability and allow other apps, reports, users and
    # analyses to automatically find figures and use them.
    #
    # Labels are case insensitive, and accept a-z, 0-9, and hyphens which can be used literally in search and are
    # also used to separate words in natural language search.
    analysis.output_manifest.get_dataset("cleaned_met_mast_data").labels = ["met", "mast", "cleaned"]

    # Finalise the analysis. This validates the output data and output manifest against the twine and optionally
    # uploads any datasets in the output manifest to the service's cloud bucket. Signed URLs are provided so that
    # the parent that asked the service for the analysis can access the data (until the signed URLs expire).
    analysis.finalise(upload_output_datasets_to=f"gs://{TEST_BUCKET_NAME}/output/test_using_manifests_analysis")

    # We're done! There's only one datafile in the output dataset, but you could create thousands more and add them
    # all :)
    #
    # If you're running this on your local machine, that's it - but when this code runs as an analysis in the cloud,
    # the files in the output manifest are copied into the cloud store. Their names and labels can be registered in a
    # search index so your colleagues can find the dataset you've produced.



import functools
import itertools


def get_nested_attribute(instance, nested_attribute_name):
    """Get the value of a nested attribute from a class instance or dictionary, with each level of nesting being
    another dictionary or class instance.

    :param dict|object instance:
    :param str nested_attribute_names: dot-separated nested attribute name e.g. "a.b.c", "a.b", or "a"
    :return any:
    """
    nested_attribute_names = nested_attribute_name.split(".")
    return functools.reduce(getattr_or_subscribe, nested_attribute_names, instance)


def has_nested_attribute(instance, nested_attribute_name):
    """Check if a class instance or dictionary has a nested attribute with the given name (each level of nesting being
    another dictionary or class instance).

    :param dict|object instance:
    :param str nested_attribute_names: dot-separated nested attribute name e.g. "a.b.c", "a.b", or "a"
    :return bool:
    """
    try:
        get_nested_attribute(instance, nested_attribute_name)
    except AttributeError:
        return False
    return True


def getattr_or_subscribe(instance, name):
    """Get an attribute from a class instance or a value from a dictionary.

    :param dict|object instance:
    :param str name: name of attribute or dictionary key
    :return any:
    """
    try:
        return getattr(instance, name)
    except AttributeError:
        try:
            return instance[name]
        except (TypeError, KeyError):
            raise AttributeError(f"{instance!r} does not have an attribute or key named {name!r}.")


def dictionary_product(keep_none_values=False, **kwargs):
    """Yield from the Cartesian product of the kwargs' iterables, optionally including `None`-valued kwargs. This
    function acts like `itertools.product` but yields each combination as a dictionary instead of a tuple. The
    dictionaries have the kwargs' keys as their keys.

    :param bool keep_none_values: if `True`, include `None`-valued kwargs in the yielded combinations
    :param kwargs: any number of iterables
    :yield dict: one of the combinations of the kwargs' iterables as a dictionary
    """
    iterables = {}
    nones = {}

    # Put kwargs with `None` values in a separate dictionary to kwargs with non-`None` values.
    for key, value in kwargs.items():
        if value is None:
            nones[key] = value
        else:
            iterables[key] = value

    # Yield each possible combination from the Cartesian product of the kwargs' iterables.
    for combination in itertools.product(*iterables.values()):
        combination_dict = dict(zip(iterables.keys(), combination))

        # Optionally add the `None`-valued kwargs to what's yielded.
        if keep_none_values:
            combination_dict = {**combination_dict, **nones}

        yield combination_dict



import logging

from octue.cloud.events.replayer import EventReplayer


logger = logging.getLogger(__name__)


class ChildEmulator:
    """An emulator for the `octue.resources.child.Child` class that handles the given events without contacting the real
    child or using Pub/Sub. Any events a real child could produce are supported. `Child` instances can be
    replaced/mocked like-for-like by `ChildEmulator` without the parent knowing.

    :param list(dict(dict))|None events: the list of events to send to the parent; each event must have an "event" key and an "attributes" key, and all events must conform to the service communication schema
    :param kwargs: any number of keyword arguments that would normally be passed to `Child.__init__`
    :return None:
    """

    def __init__(self, events=None, **kwargs):
        self.events = events or []

        if len(self.events) == 0:
            raise ValueError(
                f"A non-zero number of events must be provided to the child emulator - received {self.events!r}."
            )

        self.id = self.events[0].get("attributes", {}).get("sender")

    def __repr__(self):
        """Represent a child emulator as a string.

        :return str:
        """
        return f"<{type(self).__name__}({self.id!r})>"

    @property
    def received_events(self):
        """Get the events received from the child.

        :return list(dict):
        """
        return self.events

    def ask(self, handle_monitor_message=None, record_events=True, asynchronous=False, **kwargs):
        """Ask the child emulator a question and receive its emulated response events. Unlike a real child, the input
         values and manifest are not validated against the schema in the child's twine as it is only available to the
         real child. Hence, the input values and manifest do not affect the events returned by the emulator.

        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param bool record_events: if `True`, record events received from the child in the `received_events` property
        :param bool asynchronous: if `True`, don't wait for an answer or create an answer subscription (the result and other events can be retrieved from the event store later)
        :param kwargs: any number of keyword arguments that would normally be passed to `Child.ask`
        :return (dict, str)|(None, str): a dictionary containing the keys "output_values" and "output_manifest" (or `None` if the question is asynchronous), and the question UUID
        """
        question_uuid = self.events[0].get("attributes", {}).get("question_uuid")

        if asynchronous:
            return (None, question_uuid)

        event_replayer = EventReplayer(handle_monitor_message=handle_monitor_message, record_events=record_events)
        result = event_replayer.handle_events(self.events)
        return (result, question_uuid)



import numpy


def mandelbrot(width, height, x_range, y_range, n_iterations, c=None):
    """Compute the heightmap for a mandelbrot or julia set fractal.

    Each iteration, the new value of a pixel is calculated as z = z^2 + c, where for a mandelbrot set (default)
    c is a function of the position in the complex plane (c = x + iy) or a optionally for a julia set, c is a constant

    :param int width: Integer width of the final fractal image in pixels
    :param int height: Integer height of the final fractal image in pixels
    :param list(float)|numpy.ndarray x_range: The range of x [min, max] for which the fractal will be drawn
    :param list(float)|numpy.ndarray y_range: The range of y [min, max] for which the fractal will be drawn
    :param int n_iterations: the number of iterations limit used to compute the fractal
    :param None|(float, float) c: Optional 2-tuple (or other iterable) containing real and complex parts of constant coefficient c. Giving this argument will result in creation of a julia set, not the default mandelbrot set
    :return (numpy.ndarray, numpy.ndarray, numpy.ndarray): x, y, z values of pixel locations in the x, y complex plane and a corresponding heightmap z, with which you can lot a fancy looking 3d fractal
    """
    # Create a linearly spaced 2d grid
    [x, y] = numpy.meshgrid(
        numpy.linspace(x_range[0], x_range[1], width), numpy.linspace(y_range[0], y_range[1], height)
    )

    # Preallocate output array
    z = numpy.zeros((height, width))

    # Simple loop to render the fractal set. This is not efficient python and would be vectorised in production, but the
    # purpose here is just to provide a simple demo.
    for idx, a in numpy.ndenumerate(x):

        # Get default constant c (Mandelbrot sets use spatial coordinates as constants)
        if c is None:
            b = y[idx]
            c = (a, b)

        x_old = 0.0
        y_old = 0.0
        k = 1
        while (k <= n_iterations) and ((x_old**2.0 + y_old**2.0) < 4.0):
            x_new = x_old**2.0 - y_old**2.0 + c[0]
            y_new = 2.0 * x_old * y_old + c[1]
            x_old = x_new
            y_old = y_new
            k += 1

        z[idx] = k

    return x, y, z



import copy
import functools
import logging
import logging.handlers
import os
import re
import uuid

import google.api_core.exceptions
from google import auth
from google.cloud import secretmanager
from jsonschema import ValidationError, validate as jsonschema_validate

import twined.exceptions
from octue import exceptions
from octue.app_loading import AppFrom
from octue.diagnostics import Diagnostics
from octue.log_handlers import AnalysisLogFormatterSwitcher
from octue.resources import Child
from octue.resources.analysis import CLASS_MAP, Analysis
from octue.resources.datafile import downloaded_files
from octue.utils.files import registered_temporary_directories
from twined import Twine


SAVE_DIAGNOSTICS_OFF = "SAVE_DIAGNOSTICS_OFF"
SAVE_DIAGNOSTICS_ON_CRASH = "SAVE_DIAGNOSTICS_ON_CRASH"
SAVE_DIAGNOSTICS_ON = "SAVE_DIAGNOSTICS_ON"
SAVE_DIAGNOSTICS_MODES = {SAVE_DIAGNOSTICS_OFF, SAVE_DIAGNOSTICS_ON_CRASH, SAVE_DIAGNOSTICS_ON}


logger = logging.getLogger(__name__)


class Runner:
    """A runner of analyses for a given service.

    The ``Runner`` class provides a set of configuration parameters for use by your application, together with a range
    of methods for managing input and output file parsing as well as controlling logging.

    :param callable|type|module|str app_src: either a function that accepts an Octue analysis, a class with a ``run`` method that accepts an Octue analysis, or a path to a directory containing an ``app.py`` file containing one of these
    :param str|dict|twined.Twine twine: path to the twine file, a string containing valid twine json, or a Twine instance
    :param str|dict|None configuration_values: The strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param str|dict|None configuration_manifest: The strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param str|list(dict)|None children: The children strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param str|None output_location: the path to a cloud directory to save output datasets at
    :param str|None diagnostics_cloud_path: the path to a cloud directory to store diagnostics (this includes the configuration, input values and manifest, and logs for each question)
    :param str|None project_name: name of Google Cloud project to get credentials from
    :param str|None service_id: the ID of the service being run
    :param bool delete_local_files: if `True`, delete any files downloaded and registered temporary directories created during an analysis once it's finished
    :return None:
    """

    def __init__(
        self,
        app_src,
        twine="twine.json",
        configuration_values=None,
        configuration_manifest=None,
        children=None,
        output_location=None,
        diagnostics_cloud_path=None,
        project_name=None,
        service_id=None,
        service_registries=None,
        delete_local_files=False,
    ):
        self.app_source = app_src
        self.children = children

        if output_location and not re.match(r"^gs://[a-z\d][a-z\d_./-]*$", output_location):
            raise exceptions.InvalidInputException(
                "The output location must be a Google Cloud Storage path e.g. 'gs://bucket-name/output_directory'."
            )

        self.output_location = output_location

        # Get configuration before any transformations have been applied.
        self.diagnostics = Diagnostics(cloud_path=diagnostics_cloud_path)

        self.diagnostics.add_data(
            configuration_values=configuration_values,
            configuration_manifest=configuration_manifest,
        )

        # Ensure the twine is present and instantiate it.
        if isinstance(twine, Twine):
            self.twine = twine
        else:
            self.twine = Twine(source=twine)

        logger.debug("Parsed twine with strands %r", self.twine.available_strands)

        # Validate and initialise configuration data.
        self.configuration = self.twine.validate(
            configuration_values=configuration_values,
            configuration_manifest=configuration_manifest,
            cls=CLASS_MAP,
        )
        logger.debug("Configuration validated.")

        self.service_id = service_id
        self.service_registries = service_registries
        self.delete_local_files = delete_local_files
        self._project_name = project_name

    @classmethod
    def from_configuration(
        cls,
        service_configuration,
        app_configuration,
        project_name=None,
        service_id=None,
        **overrides,
    ):
        """Instantiate a runner from a service and app configuration.

        :param octue.configuration.ServiceConfiguration service_configuration:
        :param octue.configuration.AppConfiguration app_configuration:
        :param str|None project_name: name of Google Cloud project to get credentials from
        :param str|None service_id: the ID of the service being run
        :param overrides: optional keyword arguments to override the `Runner` instantiation parameters extracted from the service and app configuration
        :return octue.runner.Runner: a runner configured with the given service and app configuration
        """
        inputs = {
            "app_src": service_configuration.app_source_path,
            "twine": service_configuration.twine_path,
            "configuration_values": app_configuration.configuration_values,
            "configuration_manifest": app_configuration.configuration_manifest,
            "children": app_configuration.children,
            "output_location": app_configuration.output_location,
            "diagnostics_cloud_path": service_configuration.diagnostics_cloud_path,
            "project_name": project_name,
            "service_id": service_id,
            "service_registries": service_configuration.service_registries,
            "delete_local_files": service_configuration.delete_local_files,
        }

        inputs |= overrides

        return cls(**inputs)

    def __repr__(self):
        """Represent the runner as a string.

        :return str: the runner represented as a string.
        """
        return f"<{type(self).__name__}({self.service_id!r})>"

    def run(
        self,
        analysis_id=None,
        input_values=None,
        input_manifest=None,
        children=None,
        analysis_log_level=logging.INFO,
        analysis_log_handler=None,
        handle_monitor_message=None,
        save_diagnostics=SAVE_DIAGNOSTICS_ON_CRASH,
        originator_question_uuid=None,
        originator=None,
    ):
        """Run an analysis.

        :param str|None analysis_id: UUID of analysis
        :param str|dict|None input_values: the input_values strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
        :param str|dict|octue.resources.manifest.Manifest|None input_manifest: The input_manifest strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
        :param list(dict)|None children: a list of children to use instead of the children provided at instantiation. These should be in the same format as in an app's app configuration file and have the same keys.
        :param str analysis_log_level: the level below which to ignore log messages
        :param logging.Handler|None analysis_log_handler: the logging.Handler instance which will be used to handle logs for this analysis run. Handlers can be created as per the logging cookbook https://docs.python.org/3/howto/logging-cookbook.html but should use the format defined above in LOG_FORMAT.
        :param callable|None handle_monitor_message: a function that sends monitor messages to the parent that requested the analysis
        :param str save_diagnostics: must be one of {"SAVE_DIAGNOSTICS_OFF", "SAVE_DIAGNOSTICS_ON_CRASH", "SAVE_DIAGNOSTICS_ON"}; if turned on, allow the input values and manifest (and its datasets) to be saved either all the time or just if the analysis fails
        :param str|None originator_question_uuid: the UUID of the question that triggered all ancestor questions of this analysis; if `None`, this question is assumed to be the originator question
        :param str|None originator: the SRUID of the service revision that triggered all ancestor questions of this question;  if `None`, this service revision is assumed to be the originator
        :return octue.resources.analysis.Analysis:
        """
        if save_diagnostics not in SAVE_DIAGNOSTICS_MODES:
            raise ValueError(
                f"`save_diagnostics` must be one of {SAVE_DIAGNOSTICS_MODES!r}; received {save_diagnostics!r}."
            )

        # Set the analysis ID if one isn't given.
        analysis_id = str(analysis_id) if analysis_id else str(uuid.uuid4())

        # This analysis is the parent question.
        parent_question_uuid = analysis_id

        # If the originator question UUID isn't provided, assume that this analysis is the originator question.
        originator_question_uuid = originator_question_uuid or analysis_id

        # Get inputs before any transformations have been applied.
        self.diagnostics.add_data(
            analysis_id=analysis_id,
            input_values=input_values,
            input_manifest=input_manifest,
        )

        if hasattr(self.twine, "credentials"):
            self._populate_environment_with_google_cloud_secrets()
            credentials = self.twine.credentials
        else:
            credentials = None

        inputs = self.twine.validate(
            input_values=input_values,
            input_manifest=input_manifest,
            credentials=credentials,
            children=children or self.children,
            cls=CLASS_MAP,
            allow_missing=False,
            allow_extra=False,
        )
        logger.debug("Inputs validated.")

        inputs_and_configuration = {**self.configuration, **inputs}

        for manifest_strand in self.twine.available_manifest_strands:
            if manifest_strand == "output_manifest":
                continue

            self._validate_dataset_file_tags(
                manifest_kind=manifest_strand,
                manifest=inputs_and_configuration[manifest_strand],
            )

        if inputs["children"] is not None:
            inputs["children"] = self._instantiate_children(
                serialised_children=inputs["children"],
                parent_question_uuid=parent_question_uuid,
                originator_question_uuid=originator_question_uuid,
                originator=originator,
            )

        outputs_and_monitors = self.twine.prepare("monitor_message", "output_values", "output_manifest", cls=CLASS_MAP)

        if analysis_log_handler:
            extra_log_handlers = [analysis_log_handler]
        else:
            extra_log_handlers = []

        # Temporarily replace the root logger's handlers with a `StreamHandler` and the analysis log handler that
        # include the analysis ID in the logging metadata.
        with AnalysisLogFormatterSwitcher(
            analysis_id=analysis_id,
            logger=logging.getLogger(),
            analysis_log_level=analysis_log_level,
            extra_log_handlers=extra_log_handlers,
        ):

            analysis = Analysis(
                id=analysis_id,
                twine=self.twine,
                handle_monitor_message=handle_monitor_message,
                output_location=self.output_location,
                **self.configuration,
                **inputs,
                **outputs_and_monitors,
            )

            try:
                self._load_and_run_app(analysis)

            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(f"{e.msg} in {os.path.abspath(self.app_source)!r}.")

            except Exception as analysis_error:
                logger.warning("App failed.")

                if save_diagnostics in {SAVE_DIAGNOSTICS_ON_CRASH, SAVE_DIAGNOSTICS_ON}:
                    self.diagnostics.upload()

                raise analysis_error

            finally:
                for i, thread in enumerate(analysis._periodic_monitor_message_sender_threads):
                    thread.cancel()
                    logger.debug("Periodic monitor message thread %d stopped.", i)

            self._finalise_and_clean_up(analysis, save_diagnostics)
            return analysis

    def _populate_environment_with_google_cloud_secrets(self):
        """Get any secrets specified in the credentials strand from Google Cloud Secret Manager and put them in the
        local environment, ready for use by the runner.

        :return None:
        """
        missing_credentials = tuple(
            credential for credential in self.twine.credentials if credential["name"] not in os.environ
        )

        if not missing_credentials:
            return

        google_cloud_credentials, project_name = auth.default()
        secrets_client = secretmanager.SecretManagerServiceClient(credentials=google_cloud_credentials)

        if google_cloud_credentials is None:
            project_name = self._project_name

        for credential in missing_credentials:
            secret_path = secrets_client.secret_version_path(
                project=project_name,
                secret=credential["name"],
                secret_version="latest",
            )

            try:
                secret = secrets_client.access_secret_version(name=secret_path).payload.data.decode("UTF-8")
            except google.api_core.exceptions.NotFound:
                # No need to raise an error here as the Twine validation that follows will do so.
                continue

            os.environ[credential["name"]] = secret

    def _validate_dataset_file_tags(self, manifest_kind, manifest):
        """Validate the tags of the files of each dataset in the manifest against the file tags template in the
        corresponding dataset field in the given manifest field of the twine.

        :param str manifest_kind: the kind of manifest that's being validated (so the correct schema can be accessed)
        :param octue.resources.manifest.Manifest manifest: the manifest whose datasets' files are to be validated
        :return None:
        """
        # This is the manifest schema included in the twine.json file, not the schema for `manifest.json` files.
        manifest_schema = getattr(self.twine, manifest_kind)

        for dataset_name, dataset_schema in manifest_schema["datasets"].items():
            dataset = manifest.datasets.get(dataset_name)
            file_tags_template = dataset_schema.get("file_tags_template")

            # Allow optional datasets in future (not currently allowed by `twined`).
            if not (dataset and file_tags_template):
                continue

            for file in dataset.files:
                try:
                    jsonschema_validate(instance=dict(file.tags), schema=file_tags_template)
                except ValidationError as e:
                    message = (
                        e.message + f" for files in the {dataset_name!r} dataset. The affected datafile is "
                        f"{file.name!r}. Add the property to the datafile as a tag to fix this."
                    )

                    raise twined.exceptions.invalid_contents_map[manifest_kind](message)

    def _instantiate_children(self, serialised_children, parent_question_uuid, originator_question_uuid, originator):
        """Instantiate children from their serialised form (e.g. as given in the app configuration) so they are ready
        to be asked questions. Two sets of modifications are made to each child's `ask` method:

        1. The parent question UUID is set to the current analysis ID, and the originator question UUID and originator
           are set.

        2. For diagnostics, the `ask` method is wrapped so the runner can record the questions asked by the app, the
           responses received to each question, and the order the questions are asked in.

        :param list(dict) serialised_children: serialised children from e.g. the app configuration file
        :param str|None parent_question_uuid: the UUID of the question that triggered this analysis
        :param str originator_question_uuid: the UUID of the question that triggered all ancestor questions of this analysis
        :param str originator: the SRUID of the service revision that triggered the tree of questions this analysis is related to
        :return dict: a mapping of child keys to `octue.resources.child.Child` instances
        """
        children = {}

        for uninstantiated_child in serialised_children:
            child = Child(
                id=uninstantiated_child["id"],
                backend=uninstantiated_child["backend"],
                internal_sruid=self.service_id,
                service_registries=self.service_registries,
            )

            child.ask = self._add_child_question_and_response_recording(child, uninstantiated_child["key"])

            child.ask = functools.partial(
                child.ask,
                parent_question_uuid=parent_question_uuid,
                originator_question_uuid=originator_question_uuid,
                originator=originator,
            )

            children[uninstantiated_child["key"]] = child

        return children

    def _add_child_question_and_response_recording(self, child, key):
        """Add question and response recording to the `ask` method of the given child. This allows the runner to record
        the questions asked by the app, the responses received to each question, and the order the questions are asked
        in for diagnostics.

        :param octue.resources.child.Child child: the child to add question and response recording to
        :param str key: the key used to identify the child within the service
        :return callable: the wrapped `Child.ask` method
        """
        # Copy the `ask` method to avoid an infinite recursion.
        original_ask_method = copy.copy(child.ask)

        def wrapper(*args, **kwargs):
            # Convert args to kwargs so all inputs to the `ask` method can be recorded whether they're provided
            # positionally or as keyword arguments.
            kwargs.update(dict(zip(original_ask_method.__func__.__code__.co_varnames, args)))

            try:
                return original_ask_method(**kwargs)
            finally:
                self.diagnostics.add_question({"id": child.id, "key": key, **kwargs, "events": child.received_events})

        return wrapper

    def _load_and_run_app(self, analysis):
        """Load and run the app on the given analysis object.

        :param octue.resources.analysis.Analysis analysis: the analysis object containing the configuration and inputs to run the app on
        :return None:
        """
        # App as a class that takes "analysis" as a constructor argument and contains a method named "run" that
        # takes no arguments.
        if isinstance(self.app_source, type):
            self.app_source(analysis).run()
            return

        # App as a module containing a function named "run" that takes "analysis" as an argument.
        if hasattr(self.app_source, "run"):
            self.app_source.run(analysis)
            return

        # App as a string path to a module containing a class named "App" or a function named "run". The same
        # other specifications apply as described above.
        if isinstance(self.app_source, str):
            with AppFrom(self.app_source) as app:
                if hasattr(app.app_module, "App"):
                    app.app_module.App(analysis).run()
                else:
                    app.app_module.run(analysis)
            return

        # App as a function that takes "analysis" as an argument.
        self.app_source(analysis)

    def _finalise_and_clean_up(self, analysis, save_diagnostics):
        """Do the following:

        1. Finalise the analysis
        2. If diagnostics are switched on, upload the diagnostics
        3. If `delete_local_files=True`, delete any datafiles downloaded and registered temporary directories created during the analysis

        :param octue.resources.analysis.Analysis analysis: the analysis object containing the configuration and inputs to run the app on
        :param str save_diagnostics: must be one of {"SAVE_DIAGNOSTICS_OFF", "SAVE_DIAGNOSTICS_ON_CRASH", "SAVE_DIAGNOSTICS_ON"}; if turned on, allow the input values and manifest (and its datasets) to be saved either all the time or just if the analysis fails
        :return None:
        """
        if not analysis.finalised:
            analysis.finalise()

        if save_diagnostics == SAVE_DIAGNOSTICS_ON:
            self.diagnostics.upload()

        if self.delete_local_files:
            # Delete temporary directories first as this will delete entire downloaded datasets.
            if registered_temporary_directories:
                logger.warning(
                    "Deleting registered temporary directories created during analysis. This is not thread-safe - set "
                    "`delete_local_files=False` at instantiation of `Runner` to switch this off."
                )

                for dir in registered_temporary_directories:
                    logger.debug("Deleting temporary directory at %r.", dir.name)
                    dir.cleanup()

            # Then delete any datafiles were downloaded separately from a dataset.
            if downloaded_files:
                logger.warning(
                    "Deleting datafiles downloaded during analysis. This is not thread-safe - set "
                    "`delete_local_files=False` at instantiation of `Runner` to switch this off."
                )

                for path in downloaded_files:
                    if not os.path.exists(path):
                        continue

                    logger.debug("Deleting downloaded file at %r.", path)

                    try:
                        os.remove(path)
                    except FileNotFoundError:
                        logger.debug("Couldn't delete %r - it was already deleted.", path)



import logging
import os
import re

import coolname
import requests

import octue.exceptions


logger = logging.getLogger(__name__)

SERVICE_NAMESPACE_AND_NAME_PATTERN = r"([a-z0-9])+(-([a-z0-9])+)*"
COMPILED_SERVICE_NAMESPACE_AND_NAME_PATTERN = re.compile(SERVICE_NAMESPACE_AND_NAME_PATTERN)

REVISION_TAG_PATTERN = r"([A-z0-9_])+([-.]*([A-z0-9_])+)*"
COMPILED_REVISION_TAG_PATTERN = re.compile(REVISION_TAG_PATTERN)

SRUID_PATTERN = rf"^{SERVICE_NAMESPACE_AND_NAME_PATTERN}\/{SERVICE_NAMESPACE_AND_NAME_PATTERN}:{REVISION_TAG_PATTERN}$"
COMPILED_SRUID_PATTERN = re.compile(SRUID_PATTERN)


def get_sruid_parts(service_configuration):
    """Get the namespace, name, and revision tag for the service from either the service environment variables or the
    service configuration (in that order of precedence). The service revision tag is `None` if it's not provided in the
    `OCTUE_SERVICE_REVISION_TAG` environment variable as it can't be specified in the service configuration.

    :param octue.configuration.ServiceConfiguration service_configuration: the service configuration to get the service namespace and name from
    :return (str, str, str|None):
    """
    service_namespace = os.environ.get("OCTUE_SERVICE_NAMESPACE")
    service_name = os.environ.get("OCTUE_SERVICE_NAME")
    service_revision_tag = os.environ.get("OCTUE_SERVICE_REVISION_TAG")

    if service_namespace and service_namespace != service_configuration.namespace:
        logger.warning(
            "The namespace in the service configuration %r has been overridden by the `OCTUE_SERVICE_NAMESPACE` "
            "environment variable %r.",
            service_configuration.namespace,
            service_namespace,
        )
    else:
        service_namespace = service_configuration.namespace

    if service_name and service_name != service_configuration.name:
        logger.warning(
            "The name in the service configuration %r has been overridden by the `OCTUE_SERVICE_NAME` environment "
            "variable %r.",
            service_configuration.name,
            service_name,
        )
    else:
        service_name = service_configuration.name

    if service_revision_tag:
        logger.info(
            "Service revision tag %r provided by `OCTUE_SERVICE_REVISION_TAG` environment variable.",
            service_revision_tag,
        )

    return service_namespace, service_name, service_revision_tag


def create_sruid(namespace, name, revision_tag=None):
    """Create and validate a service revision unique identifier (SRUID) from a namespace, name, and revision tag. If no
    revision tag is given, a "cool name" revision tag is generated.

    :param str namespace: the name of the group to which the service belongs
    :param str name: the name of the service
    :param str|None revision_tag: a tag that uniquely identifies a particular revision of the service
    :raise octue.exceptions.InvalidServiceID: if any of the namespace, name, or revision tag are invalid
    :return str: the valid SRUID comprising the namespace, name, and revision tag
    """
    revision_tag = revision_tag or coolname.generate_slug(2)
    validate_sruid(namespace=namespace, name=name, revision_tag=revision_tag)
    return f"{namespace}/{name}:{revision_tag}"


def validate_sruid(sruid=None, namespace=None, name=None, revision_tag=None):
    """Raise an error if the service revision unique identifier (SRUID) or its components don't meet the required
    patterns. Either the `service_id` or all of the `namespace`, `name`, and `revision_tag` arguments must be given.

    :param str|None sruid: the service SRUID to validate
    :param str|None namespace: the namespace of a service to validate
    :param str|None name: the name of a service to validate
    :param str|None revision_tag: the revision tag of a service to validate
    :raise octue.exceptions.InvalidServiceID: if the service SRUID or any of its components are invalid
    :return None:
    """
    if sruid:
        if not COMPILED_SRUID_PATTERN.fullmatch(sruid):
            raise octue.exceptions.InvalidServiceID(
                f"{sruid!r} is not a valid service revision unique identifier (SRUID). It must be in the format "
                f"<namespace>/<name>:<revision_tag>. The namespace and name must be lower kebab case (i.e. only "
                f"contain the letters [a-z], numbers [0-9], and hyphens [-]) and not begin or end with a hyphen. The "
                f"revision tag can contain lowercase and uppercase letters, numbers, underscores, periods, and "
                f"hyphens, but can't start with a period or a dash. It can contain a maximum of 128 characters. These "
                f"requirements are the same as the Docker tag format."
            )

        revision_tag = sruid.split(":")[-1]

        if len(revision_tag) > 128:
            raise octue.exceptions.InvalidServiceID(
                f"The maximum length for a revision tag is 128 characters. Received {revision_tag!r}."
            )

        return

    if any((namespace is None, name is None, revision_tag is None)):
        raise ValueError(
            "If not providing the `service_id` argument for SRUID validation, all of the `namespace`, `name`, and "
            "`revision_tag` arguments must be provided instead."
        )

    validate_namespace(namespace)
    validate_name(name)

    if len(revision_tag) > 128:
        raise octue.exceptions.InvalidServiceID(
            f"The maximum length for a revision tag is 128 characters. Received {revision_tag!r}."
        )

    if not COMPILED_REVISION_TAG_PATTERN.fullmatch(revision_tag):
        raise octue.exceptions.InvalidServiceID(
            f"{revision_tag!r} is not a valid revision tag for a service. It can contain lowercase and uppercase "
            "letters, numbers, underscores, periods, and hyphens, but can't start with a period or a dash. It can "
            "contain a maximum of 128 characters. These requirements are the same as the Docker tag format."
        )


def validate_service_id(service_id=None, namespace=None, name=None):
    """Raise an error if the service ID or its components don't meet the required patterns. Either the `service_id` or
    both the `namespace` and `name` arguments must be given.

    :param str|None service_id: the service ID to validate
    :param str|None namespace: the namespace of a service to validate
    :param str|None name: the name of a service to validate
    :raise octue.exceptions.InvalidServiceID: if the service ID or any of its components are invalid
    :return None:
    """
    if service_id:
        if not COMPILED_SERVICE_NAMESPACE_AND_NAME_PATTERN.fullmatch(service_id):
            raise octue.exceptions.InvalidServiceID(
                f"{service_id!r} is not a valid service ID. It must be in the format <namespace>/<name>. The namespace "
                "and name must be lower kebab case (i.e. only contain the letters [a-z], numbers [0-9], and hyphens [-]"
                ") and not begin or end with a hyphen."
            )

        return

    if any((namespace is None, name is None)):
        raise ValueError(
            "If not providing the `service_id` argument for service ID validation, both the `namespace` and `name` "
            "arguments must be provided instead."
        )

    validate_namespace(namespace)
    validate_name(name)


def validate_namespace(namespace):
    """Raise an error if the service namespace doesn't meet the required patterns.

    :param str|None namespace: the namespace to validate
    :return None:
    """
    if not COMPILED_SERVICE_NAMESPACE_AND_NAME_PATTERN.fullmatch(namespace):
        raise octue.exceptions.InvalidServiceID(
            f"{namespace!r} is not a valid namespace for a service. It must be lower kebab case (i.e. only contain "
            "the letters [a-z], numbers [0-9], and hyphens [-]) and not begin or end with a hyphen."
        )


def validate_name(name):
    """Raise an error if the service name doesn't meet the required patterns.

    :param str|None name: the name to validate
    :return None:
    """
    if not COMPILED_SERVICE_NAMESPACE_AND_NAME_PATTERN.fullmatch(name):
        raise octue.exceptions.InvalidServiceID(
            f"{name!r} is not a valid name for a service. It must be lower kebab case (i.e. only contain the letters "
            f"[a-z], numbers [0-9], and hyphens [-]) and not begin or end with a hyphen."
        )


def convert_service_id_to_pub_sub_form(service_id):
    """Convert the service ID to the form required for use in Google Pub/Sub topic and subscription paths. This is done
    by replacing forward slashes and colons with periods and, if a service revision tag is included, replacing any
    periods in it with dashes.

    :param str service_id: a service ID or service revision unique identifier (SRUID)
    :return str: the service ID in Google Pub/Sub form
    """
    if ":" in service_id:
        service_id, service_revision_tag = service_id.split(":")
    else:
        service_revision_tag = None

    service_id = service_id.replace("/", ".")

    if service_revision_tag:
        service_id = service_id + "." + service_revision_tag.replace(".", "-")

    return service_id


def get_sruid_from_pub_sub_resource_name(name):
    """Get the SRUID from a Google Pub/Sub topic or subscription name. Note that any hyphens in the revision tag will
    be replaced with periods.

    :param str name: the name of the topic or subscription
    :return str: the SRUID of the service revision the topic or subscription is related to
    """
    namespace, name, revision_tag, *_ = name.split(".")
    return f"{namespace}/{name}:{revision_tag.replace('-', '.')}"


def split_service_id(service_id, require_revision_tag=False):
    """Split an SRUID or service ID into its namespace, name, and, if present, its revision tag. The split parts are
    validated before being returned.

    :param str service_id: the SRUID or service ID to split
    :param bool require_revision_tag: if `True`, require the service ID to include a revision tag (i.e. require the service ID to be an SRUID)
    :raise octue.exceptions.InvalidServiceID: if any of the namespace, name, or revision tag are invalid
    :return tuple(str, str, str|None): the namespace, name, and revision tag
    """
    namespace, name_and_revision_tag = service_id.split("/")

    try:
        name, revision_tag = name_and_revision_tag.split(":")
    except ValueError:
        name = name_and_revision_tag
        revision_tag = None

    if revision_tag is None and not require_revision_tag:
        validate_service_id(namespace=namespace, name=name)
    else:
        validate_sruid(namespace=namespace, name=name, revision_tag=revision_tag)

    return namespace, name, revision_tag


def get_default_sruid(namespace, name, service_registries):
    """Get the SRUID of the default revision of the service `<namespace>/<name>` if it exists in one of the specified
    service registries. The registries should be provided in priority order so that, if more than one registry contains
    a matching service, the revision that's returned is taken from the highest priority (first) registry.

    :param str namespace: the namespace of the service
    :param str name: the name of the service
    :param iter(dict) service_registries: the registries to look for the service in; the registries should be in priority order in case more than one has a service with the given namespace and name
    :raise octue.exceptions.ServiceNotFound: if a revision can't be found for the service in the service registries
    :return str: the SRUID of the default revision of the service
    """
    service_id = f"{namespace}/{name}"

    for registry in service_registries:
        response = requests.get(f"{registry['endpoint']}/{service_id}")

        if response.ok:
            revision_tag = response.json()["revision_tag"]
            logger.info("Found service revision '%s:%s' in %r registry.", service_id, revision_tag, registry["name"])
            return create_sruid(namespace=namespace, name=name, revision_tag=revision_tag)

    raise octue.exceptions.ServiceNotFound(
        f"No revisions for the service {service_id!r} were found in any of the specified service registries: "
        f"{service_registries!r}"
    )


def raise_if_revision_not_registered(sruid, service_registries):
    """Raise an error if the service revision isn't registered in the given service registries.

    :param str sruid: the SRUID of the service revision
    :param iter(dict) service_registries: the registries to look for the service revision in
    :raise octue.exceptions.ServiceNotFound: if the service revision isn't registered in any of the service registries
    :return None:
    """
    namespace, name, revision_tag = split_service_id(sruid, require_revision_tag=True)

    for registry in service_registries:
        response = requests.get(f"{registry['endpoint']}/{namespace}/{name}?revision_tag={revision_tag}")

        if response.ok:
            logger.info("Found service revision %r in %r registry.", sruid, registry["name"])
            return

    raise octue.exceptions.ServiceNotFound(
        f"Service revision {sruid!r} was not found in any of the specified service registries: {service_registries!r}"
    )



from unittest.mock import patch

from octue.cloud.emulators._pub_sub import MockSubscriber, MockSubscription, MockTopic
from octue.utils.patches import MultiPatcher


class ServicePatcher(MultiPatcher):
    """A multi-patcher that provides the patches needed to run mock services.

    :return None:
    """

    def __init__(self):
        super().__init__(
            patches=[
                patch("octue.cloud.pub_sub.service.Topic", new=MockTopic),
                patch("octue.cloud.pub_sub.service.Subscription", new=MockSubscription),
                patch("octue.cloud.pub_sub.events.SubscriberClient", new=MockSubscriber),
                patch("google.cloud.pubsub_v1.SubscriberClient", new=MockSubscriber),
            ]
        )



import abc
import importlib.metadata
import logging
import os
import re
import time
from datetime import datetime

from octue.cloud import EXCEPTIONS_MAPPING
from octue.cloud.events.validation import SERVICE_COMMUNICATION_SCHEMA, is_event_valid
from octue.definitions import GOOGLE_COMPUTE_PROVIDERS
from octue.log_handlers import COLOUR_PALETTE
from octue.resources.manifest import Manifest


logger = logging.getLogger(__name__)


if os.environ.get("COMPUTE_PROVIDER", "UNKNOWN") in GOOGLE_COMPUTE_PROVIDERS:
    # Google Cloud logs don't support colour currently - provide a no-operation function.
    colourise = lambda string, text_colour=None, background_colour=None: string
else:
    from octue.utils.colour import colourise


PARENT_SDK_VERSION = importlib.metadata.version("octue")


class AbstractEventHandler:
    """An abstract event handler for Octue service events that:
    - Provide handlers for the Octue service event kinds (see https://strands.octue.com/octue/service-communication)
    - Handles received events in the order received

    To create a concrete handler for a specific service/communication backend synchronously or asynchronously, inherit
    from this class and add the `handle_events` and `_extract_event_and_attributes` methods.

    :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive
    :param bool record_events: if `True`, record received events in the `received_events` attribute
    :param dict|None event_handlers: a mapping of event type names to callables that handle each type of event. The handlers must not mutate the events.
    :param dict schema: the JSON schema to validate events against
    :param bool include_service_metadata_in_logs: if `True`, include the SRUIDs and question UUIDs of the service revisions involved in the question to the start of the log message
    :param str|None exclude_logs_containing: if provided, skip handling log messages containing this string
    :param bool only_handle_result: if `True`, skip handling non-result events and only handle the "result" event when received (turning this on speeds up event handling)
    :param bool validate_events: if `True`, validate events before attempting to handle them (turning this off speeds up event handling)
    :return None:
    """

    def __init__(
        self,
        handle_monitor_message=None,
        record_events=True,
        event_handlers=None,
        schema=SERVICE_COMMUNICATION_SCHEMA,
        include_service_metadata_in_logs=True,
        exclude_logs_containing=None,
        only_handle_result=False,
        validate_events=True,
    ):
        self.handle_monitor_message = handle_monitor_message
        self.record_events = record_events
        self.schema = schema
        self.include_service_metadata_in_logs = include_service_metadata_in_logs
        self.exclude_logs_containing = exclude_logs_containing
        self.only_handle_result = only_handle_result
        self.validate_events = validate_events

        self.handled_events = []
        self._start_time = None

        self._event_handlers = event_handlers or {
            "delivery_acknowledgement": self._handle_delivery_acknowledgement,
            "heartbeat": self._handle_heartbeat,
            "monitor_message": self._handle_monitor_message,
            "log_record": self._handle_log_message,
            "exception": self._handle_exception,
            "result": self._handle_result,
        }

        self._log_message_colours = [COLOUR_PALETTE[1], *COLOUR_PALETTE[3:]]

    @abc.abstractmethod
    def handle_events(self, *args, **kwargs):
        """Handle events and return a handled "result" event once one is received. This method must be overridden but
        can have any arguments. The first thing it should do is call `super().handle_events()`.

        :return dict: the handled final result
        """
        self.reset()

    def reset(self):
        """Reset the handler to be ready to handle a new stream of events.

        :return None:
        """
        self._start_time = time.perf_counter()

    @abc.abstractmethod
    def _extract_event_and_attributes(self, container):
        """Extract an event and its attributes from the event container. This method must be overridden.

        :param any container: the container of the event (e.g. a Pub/Sub message)
        :return (any, dict): the event and its attributes (both must conform to the service communications event schema)
        """
        pass

    def _extract_and_validate_event(self, container):
        """Extract an event from its container, validate it, and return it.

        :param any container: the container of the event (e.g. a Pub/Sub message)
        :return (dict, dict)|(None, None): the event and its attributes if they're valid, or `(None, None)` if they're invalid
        """
        try:
            event, attributes = self._extract_event_and_attributes(container)
        except Exception:
            event = None
            attributes = {}

        # Don't assume the presence of specific attributes before validation.
        recipient = attributes.get("recipient")
        child_sdk_version = attributes.get("sender_sdk_version")

        if self.validate_events and not is_event_valid(
            event=event,
            attributes=attributes,
            recipient=recipient,
            parent_sdk_version=PARENT_SDK_VERSION,
            child_sdk_version=child_sdk_version,
            schema=self.schema,
        ):
            return (None, None)

        logger.debug(
            "%r: Received an event related to question %r.",
            attributes.get("recipient"),
            attributes.get("question_uuid"),
        )

        return (event, attributes)

    def _handle_event(self, event, attributes):
        """Pass an event to its handler and record it if appropriate.

        :param dict event: the event to handle
        :param dict attributes: the event's attributes
        :return dict|None: the output of the event (this should be `None` unless the event is a "result" event)
        """
        if self.record_events:
            self.handled_events.append({"event": event, "attributes": attributes})

        if self.only_handle_result and event["kind"] != "result":
            return

        handler = self._event_handlers[event["kind"]]
        return handler(event, attributes)

    def _handle_delivery_acknowledgement(self, event, attributes):
        """Log that the question was delivered.

        :param dict event:
        :param dict attributes: the event's attributes
        :return None:
        """
        logger.info("%r's question was delivered at %s.", attributes["recipient"], attributes["datetime"])

    def _handle_heartbeat(self, event, attributes):
        """Record the time the heartbeat was received.

        :param dict event:
        :param dict attributes: the event's attributes
        :return None:
        """
        self._last_heartbeat = datetime.now()

        logger.info(
            "%r: Received a heartbeat from service %r for question %r.",
            attributes["recipient"],
            attributes["sender"],
            attributes["question_uuid"],
        )

    def _handle_monitor_message(self, event, attributes):
        """Send the monitor message to the handler if one has been provided.

        :param dict event:
        :param dict attributes: the event's attributes
        :return None:
        """
        logger.debug(
            "%r: Received a monitor message from service %r for question %r.",
            attributes["recipient"],
            attributes["sender"],
            attributes["question_uuid"],
        )

        if self.handle_monitor_message is not None:
            self.handle_monitor_message(event["data"])

    def _handle_log_message(self, event, attributes):
        """Deserialise the event into a log record and pass it to the local log handlers. The child's SRUID and the
        question UUID are added to the start of the log message, and the SRUIDs of any subchildren called by the child
        are each coloured differently.

        :param dict event:
        :param dict attributes: the event's attributes
        :return None:
        """
        if self.exclude_logs_containing and self.exclude_logs_containing in event["log_record"]["msg"]:
            return

        record = logging.makeLogRecord(event["log_record"])

        # Split the log message into its parts.
        subchild_analysis_sections = [section.strip("[") for section in re.split("] ", record.msg)]
        final_message = subchild_analysis_sections.pop(-1)

        if self.include_service_metadata_in_logs:
            # Get information about the immediate child sending the event and colour it with the first colour in the
            # colour palette.
            immediate_child_analysis_section = colourise(
                f"[{attributes['sender']} | {attributes['question_uuid']}]",
                text_colour=self._log_message_colours[0],
            )

            # Colour any analysis sections from children of the immediate child with the rest of the colour palette.
            for i in range(len(subchild_analysis_sections)):
                subchild_analysis_sections[i] = colourise(
                    "[" + subchild_analysis_sections[i] + "]",
                    text_colour=self._log_message_colours[1:][i % len(self._log_message_colours[1:])],
                )

            record.msg = " ".join([immediate_child_analysis_section, *subchild_analysis_sections, final_message])

        else:
            record.msg = final_message

        logger.handle(record)

    def _handle_exception(self, event, attributes):
        """Raise the exception from the child.

        :param dict event:
        :param dict attributes: the event's attributes
        :raise Exception:
        :return None:
        """
        exception_message = "\n\n".join(
            (
                event["exception_message"],
                f"The following traceback was captured from the remote service {attributes['sender']!r}:",
                "".join(event["exception_traceback"]),
            )
        )

        try:
            exception_type = EXCEPTIONS_MAPPING[event["exception_type"]]

        # Allow unknown exception types to still be raised.
        except KeyError:
            exception_type = type(event["exception_type"], (Exception,), {})

        raise exception_type(exception_message)

    def _handle_result(self, event, attributes):
        """Extract any output values and output manifest from the result, deserialising the manifest if present.

        :param dict event:
        :param dict attributes: the event's attributes
        :return dict:
        """
        logger.info("%r: Received an answer to question %r.", attributes["recipient"], attributes["question_uuid"])

        if event.get("output_manifest"):
            output_manifest = Manifest.deserialise(event["output_manifest"])
        else:
            output_manifest = None

        return {"output_values": event.get("output_values"), "output_manifest": output_manifest}



import time
from threading import Timer


class RepeatingTimer(Timer):
    """A repeating version of the `threading.Timer` class."""

    def run(self):
        while not self.finished.is_set():
            time.sleep(self.interval)
            self.function(*self.args, **self.kwargs)



import copy
import json
import logging
from functools import cached_property

import coolname

from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.resources import Dataset
from octue.utils.encoders import OctueJSONEncoder


logger = logging.getLogger(__name__)


class Diagnostics:
    """A handler for question diagnostics that allows uploading of explicitly added configuration and input data and any
    questions asked to other services.

    :param str cloud_path: the cloud path of a directory to upload any added data into
    :return None:
    """

    def __init__(self, cloud_path):
        self.cloud_path = cloud_path
        self.analysis_id = None
        self.configuration_values = None
        self.configuration_manifest = None
        self.input_values = None
        self.input_manifest = None
        self.questions = []

    @cached_property
    def storage_client(self):
        """Get or instantiate the cloud storage client. No storage client is instantiated until this property is called
        for the first time. This allows checking for the `GOOGLE_APPLICATION_CREDENTIALS` environment variable to be put
        off until it's needed.

        :return octue.cloud.storage.client.GoogleCloudStorageClient:
        """
        return GoogleCloudStorageClient()

    def add_data(
        self,
        analysis_id=None,
        configuration_values=None,
        configuration_manifest=None,
        input_values=None,
        input_manifest=None,
    ):
        """Add an analysis ID, configuration values, a configuration manifest, input values, and/or an input manifest to
        the diagnostics. The values and manifests are deep-copied before being added. This method can be called
        multiple times as data becomes available. Calling again with the same keyword arguments will overwrite any data
        of that type added previously.

        :param str analysis_id: the ID of the analysis to save diagnostics for
        :param any configuration_values: configuration values to save
        :param any configuration_manifest: a configuration manifest to save
        :param any input_values: input values to save
        :param any input_manifest: an input manifest to save
        :return None:
        """
        if analysis_id:
            self.analysis_id = analysis_id

        if configuration_values:
            self.configuration_values = copy.deepcopy(configuration_values)

        if configuration_manifest:
            self.configuration_manifest = copy.deepcopy(configuration_manifest)

        if input_values:
            self.input_values = copy.deepcopy(input_values)

        if input_manifest:
            self.input_manifest = copy.deepcopy(input_manifest)

    def add_question(self, question):
        """Add a question to the list of questions to save.

        :param dict question: the question to add
        :return None:
        """
        self.questions.append(question)

    def upload(self):
        """Check that a cloud path has been provided before uploading any added data to the diagnostics cloud
        path. Any errors encountered during upload are caught and logged.

        :return None:
        """
        if not self.cloud_path:
            logger.warning(
                "Cannot upload diagnostics as the child doesn't have the `diagnostics_cloud_path` field set in its "
                "service configuration (`octue.yaml` file)."
            )
            return

        if not self.analysis_id:
            self.analysis_id = coolname.generate_slug(3)

        try:
            self._upload()
            logger.info("Diagnostics uploaded.")
        except Exception:
            logger.exception("Failed to upload diagnostics.")

    def _upload(self):
        """Upload any added data to the diagnostics cloud path.

        :return None:
        """
        question_diagnostics_path = storage.path.join(self.cloud_path, self.analysis_id)
        logger.info("Uploading diagnostics to %r.", question_diagnostics_path)

        for data_type in ("configuration", "input"):
            values_type = f"{data_type}_values"
            values = getattr(self, values_type)

            if values is not None:
                if isinstance(values, str):
                    setattr(self, values_type, self._attempt_deserialise_json(values))

                self._upload_values(values_type, question_diagnostics_path)

            manifest_type = f"{data_type}_manifest"
            manifest = getattr(self, manifest_type)

            if manifest is not None:
                if isinstance(manifest, str):
                    setattr(self, manifest_type, self._attempt_deserialise_json(manifest))

                self._upload_manifest(manifest_type, question_diagnostics_path)

        # Upload the messages received from any children.
        self.storage_client.upload_from_string(
            string=json.dumps(self.questions, cls=OctueJSONEncoder),
            cloud_path=storage.path.join(question_diagnostics_path, "questions.json"),
        )

    @staticmethod
    def _attempt_deserialise_json(string):
        """Attempt to deserialise the given string from JSON. If deserialisation fails, the original string is returned.

        :param str string: the string to attempt to deserialise
        :return any: the deserialised python object or the original string
        """
        try:
            return json.loads(string)
        except json.decoder.JSONDecodeError:
            return string

    def _upload_values(self, values_type, question_diagnostics_path):
        """Upload the values of the given type as part of the diagnostics.

        :param str values_type: one of "configuration_values" or "input_values"
        :param str question_diagnostics_path: the path to a cloud directory to upload the values into
        :return None:
        """
        values = getattr(self, values_type)

        self.storage_client.upload_from_string(
            json.dumps(values, cls=OctueJSONEncoder),
            cloud_path=storage.path.join(question_diagnostics_path, f"{values_type}.json"),
        )

    def _upload_manifest(self, manifest_type, question_diagnostics_path):
        """Upload the serialised manifest of the given type and its datasets as part of the diagnostics.

        :param str manifest_type: one of "configuration_manifest" or "input_manifest"
        :param str question_diagnostics_path: the path to a cloud directory to upload the manifest into
        :return None:
        """
        manifest = getattr(self, manifest_type)

        # Upload each dataset and update its path in the manifest.
        for dataset_name, dataset_path in manifest["datasets"].items():

            # Handle manifests containing serialised datasets instead of just the datasets' paths. Datasets can be in
            # this state when serialised if they were instantiated using the `files` argument.
            if isinstance(dataset_path, dict):
                dataset_path = dataset_path["path"]

            new_dataset_path = storage.path.join(
                question_diagnostics_path,
                f"{manifest_type}_datasets",
                dataset_name,
            )

            Dataset(dataset_path).upload(new_dataset_path)
            manifest["datasets"][dataset_name] = new_dataset_path

        # Upload the serialised manifest.
        self.storage_client.upload_from_string(
            json.dumps(manifest, cls=OctueJSONEncoder),
            cloud_path=storage.path.join(question_diagnostics_path, f"{manifest_type}.json"),
        )



import importlib.metadata
import json
import logging
from collections import defaultdict

import google.api_core

from octue.cloud.pub_sub import Subscription, Topic
from octue.cloud.pub_sub.service import PARENT_SENDER_TYPE, Service
from octue.resources import Manifest
from octue.utils.dictionaries import make_minimal_dictionary
from octue.utils.encoders import OctueJSONEncoder


logger = logging.getLogger(__name__)

TOPICS = set()
SUBSCRIPTIONS = set()
MESSAGES = defaultdict(list)


class MockTopic(Topic):
    """A mock topic that registers in a global topics dictionary rather than Google Pub/Sub."""

    def create(self, allow_existing=False):
        """Register the topic in the global topics dictionary.

        :param bool allow_existing: if True, don't raise an error if the topic already exists
        :raise google.api_core.exceptions.AlreadyExists: if the topic already exists
        :return None:
        """
        if not allow_existing:
            if self.exists():
                raise google.api_core.exceptions.AlreadyExists(f"Topic {self.path!r} already exists.")

        if not self.exists():
            TOPICS.add(self.name)
            self._created = True

    def delete(self):
        """Delete the topic from the global topics dictionary.

        :return None:
        """
        try:
            TOPICS.remove(self.name)
        except KeyError:
            pass

    def exists(self, timeout=10):
        """Check if the topic exists in the global topics dictionary.

        :param float timeout:
        :return bool:
        """
        return self.name in TOPICS


class MockSubscription(Subscription):
    """A mock subscription that registers in a global dictionary rather than Google Pub/Sub."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subscriber = MockSubscriber()

    @property
    def subscriber(self):
        return self._subscriber

    def create(self, allow_existing=False):
        """Register the subscription in the global subscriptions dictionary.

        :param bool allow_existing:
        :return None:
        """
        if not allow_existing:
            if self.exists():
                raise google.api_core.exceptions.AlreadyExists(f"Subscription {self.path!r} already exists.")

        if not self.exists():
            SUBSCRIPTIONS.add(self.name)
            self._created = True

    def delete(self):
        """Delete the subscription from the global subscriptions dictionary.

        :return None:
        """
        try:
            SUBSCRIPTIONS.remove(self.name)
        except KeyError:
            pass

    def exists(self, timeout=5):
        """Check if the subscription exists in the global subscriptions dictionary.

        :param float timeout:
        :return bool:
        """
        return self.name in SUBSCRIPTIONS


class MockFuture:
    """A mock future that is returned when publishing or subscribing."""

    def __init__(self):
        self._cancelled = False
        self._returned = False

    @property
    def returned(self):
        """Check if the future has had its `result` method called.

        :return bool:
        """
        return self._returned

    @property
    def cancelled(self):
        """Check if the future has been cancelled.

        :return bool:
        """
        return self._cancelled

    def result(self, timeout=None):
        """Mark the future as returned.

        :param float timeout:
        :return None:
        """
        self._returned = True

    def cancel(self):
        """Mark the future as cancelled.

        :return None:
        """
        self._cancelled = True


class MockPublisher:
    """A mock publisher that puts messages in a global dictionary instead of Google Pub/Sub."""

    def __init__(self, *args, **kwargs):
        pass

    def publish(self, topic, data, retry=None, **attributes):
        """Put the data and attributes into a MockMessage and add it to the global messages dictionary before returning
        a MockFuture.

        :param str topic:
        :param bytes data:
        :param google.api_core.retry.Retry|None retry:
        :return MockFuture:
        """
        MESSAGES[attributes["question_uuid"]].append(MockMessage(data=data, attributes=attributes))
        return MockFuture()


class MockSubscriber:
    """A mock subscriber that gets messages from a global dictionary instead of Google Pub/Sub."""

    def __init__(self, credentials=None):
        self.closed = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.closed = True

    def subscribe(self, subscription, callback):
        """Simulate subscribing to a topic by returning a mock future.

        :param MockSubscription subscription:
        :param callable callback:
        :return MockFuture:
        """
        if self.closed:
            raise ValueError("ValueError: Cannot invoke RPC: Channel closed!")

        return MockFuture()

    def pull(self, request, timeout=None, retry=None):
        """Return a MockPullResponse containing one MockMessage wrapped in a MockMessageWrapper. The MockMessage is
        retrieved from the global messages dictionary for the subscription included in the request under the
        "subscription" key.

        :param dict request:
        :param float|None timeout:
        :param google.api_core.retry.Retry|None retry:
        :return MockPullResponse:
        """
        if self.closed:
            raise ValueError("ValueError: Cannot invoke RPC: Channel closed!")

        question_uuid = request["subscription"].split(".")[-1]

        try:
            return MockPullResponse(
                received_messages=[
                    MockMessageWrapper(message=MESSAGES[question_uuid].pop(0)),
                ]
            )

        except IndexError:
            return MockPullResponse(received_messages=[])

    def acknowledge(self, request):
        """Do nothing.

        :param google.pubsub_v1.types.pubsub.Subscription request:
        :return None:
        """
        pass

    def create_subscription(self, request):
        """Do nothing.

        :param google.pubsub_v1.types.pubsub.Subscription request:
        :return None:
        """
        pass


class MockPullResponse:
    """A mock PullResponse that is returned by `MockSubscriber.pull`.

    :param iter(MockMessageWrapper)|None received_messages: mock pub/sub messages wrapped in the MockMessageWrapper
    :return None:
    """

    def __init__(self, received_messages=None):
        self.received_messages = received_messages or []


class MockMessageWrapper:
    """A wrapper for MockMessage.

    :param MockMessage message:
    :return None:
    """

    def __init__(self, message):
        self.message = message
        self.ack_id = None

    def __repr__(self):
        """Represent the mock message wrapper as a string.

        :return str:
        """
        return f"<{type(self).__name__}(message={self.message})>"


class MockMessage:
    """A mock Pub/Sub message containing serialised data and any number of attributes.

    :param bytes data:
    :param dict|None attributes:
    :return None:
    """

    def __init__(self, data, attributes=None):
        self.data = data
        self.attributes = attributes or {}

        # Encode the attributes as they would be in a real Pub/Sub message.
        for key, value in self.attributes.items():
            if isinstance(value, bool):
                value = str(int(value))
            elif isinstance(value, (int, float)):
                value = str(value)

            self.attributes[key] = value

    @classmethod
    def from_primitive(cls, data, attributes):
        """Instantiate a mock message from data in the form of a JSON-serialisable python primitive.

        :param any data:
        :param dict attributes:
        :return MockMessage:
        """
        return cls(data=json.dumps(data, cls=OctueJSONEncoder).encode(), attributes=attributes)

    def __repr__(self):
        """Represent a mock message as a string.

        :return str: the mock message as a string
        """
        return f"<{type(self).__name__}(data={self.data!r})>"

    def ack(self):
        """Do nothing.

        :return None:
        """
        pass


class MockService(Service):
    """A mock Google Pub/Sub Service that can send and receive messages synchronously to other instances.

    :param octue.resources.service_backends.GCPPubSubBackEnd backend:
    :param str service_id:
    :param callable run_function:
    :param dict(str, MockService)|None children:
    :return None:
    """

    def __init__(self, backend, service_id=None, run_function=None, children=None, *args, **kwargs):
        super().__init__(backend, service_id, run_function, *args, **kwargs)
        self.children = children or {}
        self._publisher = MockPublisher()

    @property
    def publisher(self):
        return self._publisher

    def ask(
        self,
        service_id,
        input_values=None,
        input_manifest=None,
        children=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        save_diagnostics="SAVE_DIAGNOSTICS_ON_CRASH",
        question_uuid=None,
        parent_question_uuid=None,
        originator_question_uuid=None,
        originator=None,
        push_endpoint=None,
        asynchronous=False,
        retry_count=0,
        timeout=86400,
        parent_sdk_version=importlib.metadata.version("octue"),
    ):
        """Put the question into the messages register, register the existence of the corresponding response topic, add
        the response to the register, and return a MockFuture containing the answer subscription path.

        :param str service_id:
        :param dict|list|None input_values:
        :param octue.resources.manifest.Manifest|None input_manifest:
        :param list(dict)|None children:
        :param bool subscribe_to_logs:
        :param bool allow_local_files:
        :param str save_diagnostics:
        :param str|None question_uuid:
        :param str|None parent_question_uuid:
        :param str|None originator_question_uuid:
        :param str|None originator:
        :param str|None push_endpoint:
        :param bool asynchronous:
        :param int retry_count: the retry count of the question (this is zero if it's the first attempt at the question)
        :param float|None timeout:
        :param str parent_sdk_version:
        :return MockFuture, str:
        """
        response_subscription, question_uuid = super().ask(
            service_id=service_id,
            input_values=input_values,
            input_manifest=input_manifest,
            children=children,
            subscribe_to_logs=subscribe_to_logs,
            allow_local_files=allow_local_files,
            save_diagnostics=save_diagnostics,
            question_uuid=question_uuid,
            parent_question_uuid=parent_question_uuid,
            originator_question_uuid=originator_question_uuid,
            originator=originator,
            push_endpoint=push_endpoint,
            asynchronous=asynchronous,
            retry_count=retry_count,
            timeout=timeout,
        )

        # Delete question from messages sent to subscription so the parent doesn't pick it up as a response message. We
        # do this as subscription filtering isn't implemented in this set of mocks.
        MESSAGES[question_uuid].pop(0)

        question = make_minimal_dictionary(kind="question", input_values=input_values, children=children)

        # Ignore any errors from the answering service as they will be raised on the remote service in practice, not
        # locally as is done in this mock.
        if input_manifest is not None:
            question["input_manifest"] = input_manifest.to_primitive()

        # If the originator question UUID isn't provided, assume that this question is the originator.
        originator_question_uuid = originator_question_uuid or question_uuid

        # If the originator isn't provided, assume that this service revision is the originator.
        originator = originator or self.id

        try:
            self.children[service_id].answer(
                MockMessage.from_primitive(
                    data=question,
                    attributes={
                        "datetime": "2024-04-11T10:46:48.236064",
                        "uuid": "a9de11b1-e88f-43fa-b3a4-40a590c3443f",
                        "question_uuid": question_uuid,
                        "parent_question_uuid": parent_question_uuid,
                        "originator_question_uuid": originator_question_uuid,
                        "forward_logs": subscribe_to_logs,
                        "save_diagnostics": save_diagnostics,
                        "parent": self.id,
                        "originator": originator,
                        "sender": self.id,
                        "sender_type": PARENT_SENDER_TYPE,
                        "sender_sdk_version": parent_sdk_version,
                        "recipient": service_id,
                        "retry_count": retry_count,
                    },
                )
            )
        except Exception as e:  # noqa
            logger.exception(e)

        return response_subscription, question_uuid


class MockAnalysis:
    """A mock Analysis object with just the output strands.

    :param any output_values:
    :param octue.resources.manifest.Manifest|None output_manifest:
    :return None:
    """

    def __init__(self, output_values="Hello! It worked!", output_manifest=None):
        self.output_values = output_values
        self.output_manifest = output_manifest


class DifferentMockAnalysis:
    output_values = "This is another successful analysis."
    output_manifest = None


class MockAnalysisWithOutputManifest:
    output_values = "This is an analysis with an empty output manifest."
    output_manifest = Manifest()


class MockSubscriptionCreationResponse:
    """A mock response to creating a Google Cloud Pub/Sub subscription.

    :param google.pubsub_v1.types.pubsub.Subscription request:
    :return None:
    """

    def __init__(self, request):
        self.__dict__ = vars(request)


def get_pub_sub_resource_name(path):
    """Get the Pub/Sub resource name of the topic or subscription (e.g. "octue.services.<uuid>") from its path (e.g.
    "projects/<project-name>/topics/octue.services.<uuid>").

    :param str path:
    :return str:
    """
    return path.split("/")[-1]



class MixinBase:
    """Allows you to use any combination of mixin classes which pass unused *args and **kwargs up to their superclass,
    without encountering:
    ```
    Error
    Traceback (most recent call last):
        super().__init__(*args, **kwargs)
    TypeError: object.__init__() takes no parameters
    ```
    on initialisation.

    The case of passing extra args and kwargs to your class constructor which aren't popped by the various
    mixins is still handled safely with the same error (to prevent loss of variables and poorly defined behaviour
    later).
    """

    def __init__(self, *args, **kwargs):
        if (len(args) > 0) or (len(kwargs.keys()) > 0):
            raise TypeError("object.__init__() takes no parameters")
        super().__init__()



import octue.exceptions
import twined.exceptions
from octue.utils.exceptions import create_exceptions_mapping


EXCEPTIONS_MAPPING = create_exceptions_mapping(
    globals()["__builtins__"],
    vars(twined.exceptions),
    vars(octue.exceptions),
)
